/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubopt.h"
#include <QUB_Heap.h>
using namespace fq;

// Properties {
//     # mil config section:
//     ModelFile {...}
//     DFP { (as in MIL) }
//     Histograms {
//         BinCount = 32
//         PDFs Per Iteration = 0 | 1
//     }
//     StopFlag = 0 | 1
//     # msearch-specific section
//     Method = MIL | MIP
//     NoLoops = 0 | 1
//     BalanceLoops = 0 | 1
//     InitialK0 = 100.0
//     InitialK1 = 0.0
//     MaxRates = 20 # max two-way connections; i.e. number of lines
//     MaxExitRatesPerState = 5
//     RandomRates = 0 | 1 {
//         From = 0.1    # times initial rate
//         To = 10.0
//     }
//     TrialsPerModel = 1
//     Output = Iteration | Model | Final
//     KeepTopN = 1
//     TopN = 20
// }

// models {
//     N3112 = 5 7
//     {
//         NoLoops = 5
//     }
//     ...
// }

// MSearch Result {
//     Properties { ... }
//     ModelTrial {
//         ModelFile = 5 Initial { ... }
//         ModelFile = 5 Final   { ... }
//         LL = 123423.3322
//         LL per event = 3.8212
//         Gradient = 23423.3  32.0 -0.003 3.2e-6
//         Iterations = 12
//     }
// }

//----- Local Functions 

// wrapper for heap to sort by AIC ascending
class msearch_result{
public:
  QUB_Tree res;
  msearch_result() {}
  msearch_result(QUB_Tree result)
    : res(result)
  {}
  msearch_result(const msearch_result& other)
    : res(other.res)
  {}
  msearch_result& operator=(const msearch_result& other) {
    res = other.res;
    return *this;
  }
  bool operator<(const msearch_result& rhs) {
    // QUB_Heap.pop returns the biggest
    // but we want to remove the smallest, so reverse < and >
    return ( res["AIC"].dataAsDouble() > rhs.res["AIC"].dataAsDouble() );
  }
};

class msearch_eval_tree{
	public:
		msearch_eval_tree( std::vector<QUB_Tree>& dataSets, QUB_Tree config, QUB_Tree models,
							QTR_NodeFptr rptCB, QTR_NodeFptr rsltCB, QTR_NodeFptr pctCB,
							QTR_NodeFptr filterCB );

		QUB_Tree execute();

		QUB_Tree handleMilResult( QUB_Tree arg );
		QUB_Tree handleMilReport( QUB_Tree arg );

	private:
		void checkConstraints();  // before making milConfig, remove constraints from config["ModelFile"] and warn user that there can be no constraints
		void generateModelList(); // use config["ModelFile"] and models to fill trials and generate reorderedStates
		void disqualifyTrials();  // remove trials exceeding config["Max*Rates*"]
		void setupMIL();          // generate milConfig and milResultCB
		void setupRateTemplate(); // fill rateTemplate with the initial rates and appropriate structure to add to models
		void setupRates( QUB_Tree model, int trial ); // replace rates as spec. by trial, then randomize as appropriate
		void runOne( int trial, int retries=3 ); // and append to finalResult
		void calcScatter( QUB_Tree trialResult, QUB_Tree milResult );
		

		std::vector<QUB_Tree> dataSets;
		QUB_Tree config, models;
		QTR_NodeFptr reportCB, resultCB, percentCB, filterCB;

		int *Stopped;

		list<int> trials;
		int ntrials, itrial; // for computing pct
		int curTrial;
		bool freshTrial; // true from beginning of runOne() until first handleMilResult()
		int warnings;

		bool balanceLoops;
		double initK0, initK1;
		bool randomRates;
		double randomFrom, randomTo;
		bool reportIterations;
		bool keepTopN;
		int topN;

		std::vector<int> reorderedStates; // grouped by class in order from most to least populous
		QUB_Tree rateTemplate;

		QUB_Tree milConfig;
		QTR_NodeFptr milResultCB; // instance of msearch_milResultCB which calls handleMilResult
		QTR_NodeFptr milReportCB; // instance of msearch_milResultCB which calls handleMilReport
		QTR_NodeFptr milPercentCB; // ignores the percent

		QUB_Heap<msearch_result> topHeap;
		QUB_Tree finalResult;
	};

class msearch_milResultCB : public QTR_NodeFunc {
	public:
		msearch_milResultCB( msearch_eval_tree& msearch );

		virtual QUB_Tree NodeCall( QUB_Tree arg );
	private:
		msearch_eval_tree& msearch_eval;
	};

class msearch_milReportCB : public QTR_NodeFunc{
	public:
		msearch_milReportCB( msearch_eval_tree& msearch );

		virtual QUB_Tree NodeCall( QUB_Tree arg );
	private:
		msearch_eval_tree& msearch_eval;
	};

QUB_Tree msearch_tree( std::vector<QUB_Tree>& dataSets, QUB_Tree config, QUB_Tree models,
						QTR_NodeFptr rptCB, QTR_NodeFptr rsltCB, QTR_NodeFptr pctCB, QTR_NodeFptr filterCB );

string ModelClassOf( QUB_Tree model, std::vector<int>& reorderedStates );
void msearch_modelCount( QUB_Tree model, QUB_Tree models, int *allModelCount, 
						int *noLoopCount );

//-------------------------------------------------------------------------------------
msearch_eval_tree::msearch_eval_tree( std::vector<QUB_Tree>& datasets, QUB_Tree cfg, QUB_Tree mdls,
					QTR_NodeFptr rptCB, QTR_NodeFptr rsltCB, QTR_NodeFptr pctCB,
					QTR_NodeFptr filtrCB )
  : dataSets( datasets ), config( cfg ), models( mdls ),
    reportCB( rptCB ), resultCB( rsltCB ), percentCB( pctCB ), filterCB( filtrCB ),
	Stopped( 0 )
{
	QUB_Tree stopFlag = * (config.find("StopFlag"));
	if ( ! stopFlag.isNull() )
		Stopped = (int *) stopFlag.data();

	reportIterations = config["Output"].dataAsString() == "Iteration";

	checkConstraints();
	generateModelList();
	disqualifyTrials();
	setupMIL();

	balanceLoops = (0!=config["BalanceLoops"].dataAsInt(0));

	initK0      = config["InitialK0"].dataAsDouble( 100.0 );
	initK1      = config["InitialK1"].dataAsDouble( 0.0 );
	setupRateTemplate();

	QUB_Tree randomNode = config["RandomRates"];
	randomRates = (0!=randomNode.dataAsInt( 0 ));
	if ( randomRates ) {
		randomFrom = randomNode["From"].dataAsDouble( 0.1 );
		randomTo   = randomNode["To"].dataAsDouble( 10.0 );
	}
	
	keepTopN = config["KeepTopN"].dataAsInt( 1 );
	topN = config["TopN"].dataAsInt( 20 );

	finalResult = QUB_Tree::Create("MSearch Result");
	finalResult.appendClone( config );
}

QUB_Tree msearch_eval_tree::execute()
{
	int trialsPerModel = config["TrialsPerModel"].dataAsInt(1);

	ntrials = trials.size() * trialsPerModel;
	itrial = 0;

	QUB_Tree pctHolder = QUB_Tree::Create("");
	pctHolder.setData( QTR_TYPE_INT, 0 );
	int *pct = (int *) pctHolder.data();

	for ( list<int>::iterator trial_i = trials.begin(); trial_i != trials.end(); ++trial_i ) {
		curTrial = *trial_i;
		for ( int rep = 0; rep < trialsPerModel; ++rep ) {
			runOne( curTrial );

			++itrial;
			*pct = (100 * itrial) / ntrials;
			percentCB->NodeCall( pctHolder );

			if ( Stopped && *Stopped )
				return false;
		}
		if ( Stopped && *Stopped )
			return false;
	}
	
	if ( keepTopN ) {
	  // they are in reverse order, so insert each before the prev.
	  QUB_TreeIter resultIter = finalResult.end();
	  while ( topHeap.size() ) {
	    resultIter.insert( topHeap.first().res );
	    topHeap.pop();
	  }
	}

	return finalResult;
}

void msearch_eval_tree::checkConstraints()  // before making milConfig, remove constraints from config["ModelFile"] and warn user that there can be no constraints
{
	QUB_TreeIter constraints = config["ModelFile"]["Constraints"].children();
	if ( ! (*constraints).isNull() ) {
		QUB_Tree warning = QUB_Tree::Create("");
		warning.setData( "Discarding constraints" );
		reportCB->NodeCall( warning );

		while ( ! (*constraints).isNull() ) {
			constraints.remove();
		}
	}
}

void msearch_eval_tree::generateModelList() // use config["ModelFile"] and models to fill trials and generate reorderedStates
{
	QUB_Tree modelList = *( models.find( ModelClassOf( config["ModelFile"], reorderedStates ) ) );

	if ( config["NoLoops"].dataAsInt(0) ) {
		modelList = modelList["NoLoops"];
	}

	if ( modelList.dataCount() == 0 ) {
		// complain
	}
	else {
		std::vector<int> modelArr( modelList.dataCount() );
		modelList.readDataInto( &(modelArr[0]), 0, modelList.dataCount() - 1 );
		for ( int i=modelList.dataCount()-1; i>=0; --i )
			trials.insert( trials.begin(), modelArr[i] );
	}
}

void msearch_eval_tree::disqualifyTrials()  // remove trials exceeding config["MaxExitRates*"]
{
	int maxRates = config["MaxRates"].dataAsInt(100000); // i.e. unlimited unless specified
	int maxPerState = config["MaxExitRatesPerState"].dataAsInt(100000);

	list<int>::iterator here, next;
	for ( here = trials.begin(); here != trials.end(); here = next ) {
		// do the increment ahead of time in a temp, in case *here is erased
		next = here;
		++next;

		int nstate = reorderedStates.size();
		int numRates = 0;
		std::vector<int> numExitPerState(nstate, 0);

		int maxpath = nstate * (nstate - 1) / 2;
		int pathsel = 1 << (maxpath - 1);
		
		for ( int from=0; from<nstate; ++from )
			for ( int to=from+1; to<nstate; ++to, pathsel >>= 1 )
				if ( *here & pathsel ) {
					++numRates;
					++numExitPerState[ from ];
					++numExitPerState[ to ];
				}

		int maxExitOneState = *( max_element( numExitPerState.begin(), numExitPerState.end() ) );

		if ( (numRates > maxRates) || (maxExitOneState > maxPerState) )
			trials.erase( here );
	}
}

void msearch_eval_tree::setupMIL() // generate milConfig and milResultCB
{
	milConfig = QUB_Tree::Create("Properties");
	milConfig["Mode"].setData( "optimize" );
	milConfig["use segments"].setData( "together" );

	milConfig.appendClone( config["ModelFile"] );
	milConfig.appendClone( config["DFP"] );
	milConfig.appendClone( config["Histograms"] );

	QUB_Tree stopFlag = config["StopFlag"];
	if ( stopFlag.dataCount() == 0 )
		stopFlag.setData( QTR_TYPE_INT, 0 );
	milConfig.appendChild( stopFlag ); // the original, so msearch caller can change its value

	milResultCB = QTR_NodeFptr( new msearch_milResultCB( *this ) );
	milReportCB = QTR_NodeFptr( new msearch_milReportCB( *this ) );
	milPercentCB = QTR_NodeFptr( new QTR_NoNodeFunc );
}

void msearch_eval_tree::setupRateTemplate()
{
	rateTemplate = *( config["ModelFile"]["Rates"].find("Rate") );
	
	if ( rateTemplate.isNull() ) {
		rateTemplate = QUB_Tree::Create("Rate");
		rateTemplate["States"].setNumData( QTR_TYPE_INT, 2, 1 );
		rateTemplate["k0"].setNumData( QTR_TYPE_DOUBLE, 2, 1 );
		rateTemplate["k1"].setNumData( QTR_TYPE_DOUBLE, 2, 1 );
		rateTemplate["dk0"].setNumData( QTR_TYPE_DOUBLE, 2, 1 );
		rateTemplate["dk1"].setNumData( QTR_TYPE_DOUBLE, 2, 1 );
		rateTemplate["P"].setNumData( QTR_TYPE_INT, 2, 1 );
		rateTemplate["Q"].setNumData( QTR_TYPE_INT, 2, 1 );

		QUB_Tree PNames = rateTemplate["PNames"];
		QUB_Tree PName = PNames["PName"];
		PName.setData( "Ligand" );
		PNames.appendClone( PName ); // one for each direction

		QUB_Tree QNames = rateTemplate["QNames"];
		QUB_Tree QName = QNames["QName"];
		QName.setData( "Voltage" );
		QNames.appendClone( QName ); // one for each direction

		rateTemplate["PValue"].setData( QTR_TYPE_POINTER, 4, 2 );
		rateTemplate["QValue"].setData( QTR_TYPE_POINTER, 4, 2 );
	}
	else {
		rateTemplate = rateTemplate.clone();
	}

	double *dbls;
	int    *ints;

	dbls = (double *) rateTemplate["k0"].data();
	dbls[0] = dbls[1] = initK0;
	dbls = (double *) rateTemplate["k1"].data();
	dbls[0] = dbls[1] = initK1;
	dbls = (double *) rateTemplate["dk0"].data();
	dbls[0] = dbls[1] = 0.0;
	dbls = (double *) rateTemplate["dk1"].data();
	dbls[0] = dbls[1] = 0.0;
	ints = (int *) rateTemplate["P"].data();
	ints[0] = ints[1] = 0; // no way to specify which rates are conc.sensitive and it seems dumb to just make them all
	ints = (int *) rateTemplate["Q"].data();
	ints[0] = ints[1] = (initK1 != 0.0) ? 1 : 0;
	ints = (int *) rateTemplate["PValue"].data(); // this is a pointer to the value named by PName.  (void *) 0 is always a safe substitute
	ints[0] = ints[1] = 0;
	ints = (int *) rateTemplate["QValue"].data(); // this is a pointer to the value named by QName.  (void *) 0 is always a safe substitute
	ints[0] = ints[1] = 0;

	// States are still uninitialized
}

inline double randInRange( double from, double to )
{
	return from + rand() * (to - from) / RAND_MAX;
}

void msearch_eval_tree::setupRates( QUB_Tree model, int trial )
{
	QUB_Tree rates = model["Rates"];

	QUB_TreeIter rate_i = rates.children();
	while ( ! (*rate_i).isNull() )
		rate_i.remove();

	int nstate = reorderedStates.size();
	int maxpath = nstate * (nstate - 1) / 2;
	int pathsel = 1 << (maxpath - 1);

	for ( int from=0; from<nstate; ++from ) {
		for ( int to=from+1; to<nstate; ++to, pathsel >>= 1 ) {
			if ( trial & pathsel ) {
				QUB_Tree rate = rateTemplate.clone();
				rate_i.insert( rate );
				++rate_i;

				int *states = (int *) rate["States"].data();
				states[0] = reorderedStates[from];
				states[1] = reorderedStates[to];

				if ( randomRates ) {
					// randomize in log space to make smaller rates more likely (mil prefers starting rates that are too small)
					double fromLog = log( randomFrom );
					double toLog   = log( randomTo );

					double *k0 = (double *) rate["k0"].data();
					k0[0] *= exp( randInRange( fromLog, toLog ) );
					k0[1] *= exp( randInRange( fromLog, toLog ) );

					double *k1 = (double *) rate["k1"].data();
					k1[0] *= exp( randInRange( fromLog, toLog ) );
					k1[1] *= exp( randInRange( fromLog, toLog ) );
				}
			}
		}
	}
}

bool TooCloseToSearchLimit(QUB_Tree milResult)
{
	double searchLimit = 0.9 * milResult["Properties"]["SearchLimit"].dataAsDouble(1000.0);
	QUB_TreeMonoIter irate = milResult["Properties"]["ModelFile"]["Rates"].find("Rate");
	QUB_TreeMonoIter orate = milResult["ModelFile"]["Rates"].find("Rate");
	
	for ( ; ! (irate->isNull() || orate->isNull()); irate.nextSameName(), orate.nextSameName() ) {
		double *ik0 = (double *) (*irate)["k0"].data();
		double *ok0 = (double *) (*orate)["k0"].data();

		for ( int i=0; i<2; ++i ) {
			if ( (ok0[i] < (ik0[i] / searchLimit)) || (ok0[i] > (ik0[i] * searchLimit)) )
				return true;
		}
	}

	return false;
}

void msearch_eval_tree::runOne( int trial, int retries ) // and append to finalResult
{
	setupRates( milConfig["ModelFile"], trial );

	if ( balanceLoops ) {
		QUB_Tree balMdl = DoBalanceAllLoops( milConfig["ModelFile"] );
		QUB_TreeIter cns_ins = milConfig["ModelFile"]["Constraints"].children();
		while ( ! cns_ins->isNull() )
			cns_ins.remove();
		for (QUB_TreeMonoIter lci = balMdl["Constraints"].find("LoopBal"); ! lci->isNull(); lci.nextSameName()) {
			cns_ins.insert( lci->clone() );
			++cns_ins;
		}
	}

	if ( filterCB->NodeCall( milConfig["ModelFile"] ).isNull() )
		return;

	try {
		freshTrial = true;
		warnings = 0;

		QUB_Tree milResult;
		  mil_eval_tree milEval( dataSets, milConfig, milReportCB, milResultCB, milPercentCB );
		  milResult = milEval.execute();

		if ( randomRates && retries && TooCloseToSearchLimit(milResult) ) {
			runOne(trial, retries-1);
			return;
		}

		QUB_Tree trialResult = QUB_Tree::Create("ModelTrial");

		trialResult["ModelNumber"].setData( QTR_TYPE_INT, trial );
		
		trialResult.appendChild( milResult["AIC"] );
		trialResult.appendChild( milResult["Gradient"] );
		trialResult.appendChild( milResult["Iterations"] );
		trialResult.appendChild( milResult["ErrorCode"] );
		trialResult["Warnings"].setData( QTR_TYPE_INT, warnings );
		trialResult.appendChild( milResult["ParameterCount"] ).setName(" ParameterCount");
		trialResult.appendChild( milResult["LL"] ).setName(" LL");

		calcScatter( trialResult, milResult );

		QUB_Tree model = milConfig["ModelFile"].clone(); // leave the original in there for the next trial
		trialResult.appendChild( model );

		ostrstream initmdlname;
		initmdlname << trial << " Initial" << '\0';
		model.setData( initmdlname.str() );
		initmdlname.rdbuf()->freeze(0);

		model = milResult["ModelFile"];
		trialResult.appendChild( model );

		ostrstream finalmdlname;
		finalmdlname << trial << " Final" << '\0';
		model.setData( finalmdlname.str() );
		finalmdlname.rdbuf()->freeze(0);

		while ( true ) {
			QUB_Tree ds = *( milResult.find("DataSet") );
			if ( ds.isNull() )
				break;
			trialResult.appendChild( ds );
		}
		if ( keepTopN ) {
		  if ( ! trialResult["ErrorCode"].dataAsInt() ) {
		    topHeap.push( msearch_result(trialResult) );
		    if ( topHeap.size() > topN )
		      topHeap.pop();
		  }
		}
		else {
		  finalResult.appendChild( trialResult );
		}
	}
	catch (...) {
		_clearfp();
		_fpreset();
		
		QUB_Tree complaint = QUB_Tree::Create("");
		complaint.setData( "Unknown MIL error" );
		reportCB->NodeCall( complaint );
	}
}

template<class T>
T calcSD( std::vector<T>& v )
{
	if ( ! v.size() )
		return 0.0;

	T avg = 0.0;
	for ( std::vector<T>::iterator i = v.begin(); i != v.end(); ++i )
		avg += *i;
	avg /= v.size();

	T sd = 0.0;
	for ( std::vector<T>::iterator j = v.begin(); j != v.end(); ++j ) {
		T diff = avg - *j;
		sd += diff * diff;
	}
	sd = (T) sqrt( sd );

	return sd;
}

void msearch_eval_tree::calcScatter( QUB_Tree trialResult, QUB_Tree milResult )
{
	std::vector<double> normLLofSeg, normLLofFile;

	for ( QUB_TreeMonoIter dsi = milResult.find("DataSet"); ! (*dsi).isNull(); dsi.nextSameName() ) {
		double fileLL = 0.0;
		int fileDwells = 0;

		for ( QUB_TreeMonoIter segi = (*dsi).find("Segment"); ! (*segi).isNull(); segi.nextSameName() ) {
			double segLL = (*segi)["LL"].dataAsDouble();
			int segDwells = (*segi)["DwellCount"].dataAsInt(1);

			normLLofSeg.push_back( segLL / segDwells );

			fileLL += segLL;
			fileDwells += segDwells;
		}

		if ( fileDwells )
			normLLofFile.push_back( fileLL / fileDwells );
	}

	if ( normLLofSeg.size() > 1 )
		trialResult["normLL/seg scatter"].setData( QTR_TYPE_DOUBLE, calcSD( normLLofSeg ) / sqrt(double(normLLofSeg.size()-1)) );

	if ( normLLofFile.size() > 1 )
		trialResult["normLL/file scatter"].setData( QTR_TYPE_DOUBLE, calcSD( normLLofFile ) / sqrt(double(normLLofFile.size()-1)) );
}

QUB_Tree msearch_eval_tree::handleMilResult( QUB_Tree arg )
{
	if (   (! (*( arg.find("Final") )).isNull())
		|| reportIterations ) {

		if ( freshTrial ) {
			arg["ModelNumber"].setData( QTR_TYPE_INT, curTrial );

			for ( QUB_TreeMonoIter dsi = arg.find("DataSet"); ! (*dsi).isNull(); dsi.nextSameName() ) {
				QUB_Tree histGroup = * ( (*dsi).find("HistogramGroup") );
				if ( ! histGroup.isNull() ) {
					string title = histGroup["Title"].dataAsString();
					ostrstream ost;
					ost << title << " Model " << curTrial << '\0';
					histGroup["Title"].setData( ost.str() );
					ost.rdbuf()->freeze(0);
				}
			}
			// calc npar = # rates * 2 * ( (initialK1 != 0.0) ? 2 : 1 )
			int npar = 0;
			for ( QUB_TreeMonoIter ratei = arg["ModelFile"]["Rates"].find("Rate");
				  ! (*ratei).isNull(); ratei.nextSameName() )
				npar += 2;
			if ( initK1 != 0.0 )
				npar *= 2;
			for ( QUB_TreeMonoIter cnsi = arg["ModelFile"]["Constraints"].children();
			      ! (*cnsi).isNull(); ++cnsi )
			        --npar;
			arg["ParameterCount"].setData( QTR_TYPE_INT, npar );
		}

		arg["AIC"].setData( QTR_TYPE_DOUBLE, arg["LL"].dataAsDouble() - 2 * arg["ParameterCount"].dataAsInt() );

		freshTrial = false;

		return resultCB->NodeCall( arg );
	}
	else {
		return arg;
	}
}

QUB_Tree msearch_eval_tree::handleMilReport( QUB_Tree arg )
{
	++warnings;
	return reportCB->NodeCall( arg );
}

msearch_milResultCB::msearch_milResultCB( msearch_eval_tree& msearch )
  : msearch_eval( msearch )
{}

QUB_Tree msearch_milResultCB::NodeCall( QUB_Tree arg )
{
	return msearch_eval.handleMilResult( arg );
}

msearch_milReportCB::msearch_milReportCB( msearch_eval_tree& msearch )
  : msearch_eval( msearch )
{}

QUB_Tree msearch_milReportCB::NodeCall( QUB_Tree arg )
{
	return msearch_eval.handleMilReport( arg );
}

QUB_Tree msearch_tree( std::vector<QUB_Tree>& dataSets, QUB_Tree config, QUB_Tree models,
						QTR_NodeFptr rptCB, QTR_NodeFptr rsltCB, QTR_NodeFptr pctCB, QTR_NodeFptr filterCB )
{
	msearch_eval_tree eval( dataSets, config, models, rptCB, rsltCB, pctCB, filterCB );
	return eval.execute();
}

extern "C" MIL_API QTR_Impl * __stdcall
msearch_tree_iface( QTR_Impl *cfg, QTR_Impl **data, QTR_Impl *models,
			  QTR_INodeCFunc *rptCB, QTR_INodeCFunc *rsltCB, QTR_INodeCFunc *pctCB, QTR_INodeCFunc *filterCB )
{
	QUB_Tree cfgNode( cfg );

	std::vector<QUB_Tree> dataNodes;
	QTR_Impl **di = data;
	while ( *di )
		dataNodes.push_back( QUB_Tree( *(di++) ) );

	QUB_Tree modelsNode( models );

	QTR_NodeFptr rpt, rslt, pct, filter;
	if ( rptCB )
		rpt = QTR_NodeFptr( new QTR_NodePasFunc( rptCB ) );
	else
		rpt = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( rsltCB )
		rslt = QTR_NodeFptr( new QTR_NodePasFunc( rsltCB ) );
	else
		rslt = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( pctCB )
		pct = QTR_NodeFptr( new QTR_NodePasFunc( pctCB ) );
	else
		pct = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( filterCB )
		filter = QTR_NodeFptr( new QTR_NodePasFunc( filterCB ) );
	else
		filter = QTR_NodeFptr( new QTR_NoNodeFunc );

	QUB_Tree resultNode = msearch_tree( dataNodes, cfgNode, modelsNode, rpt, rslt, pct, filter );
	
	QTR_Impl *result = resultNode.getImpl();
	if ( result )
		QTR_INCREF( result );
	return result;
}

extern "C" PyObject * msearch_tree_Pyiface( PyObject *self, PyObject *args )
{
	PyObject *cfgImpl, *dataseq, *modelsImpl, *rptPy = 0, *rsltPy = 0, *pctPy = 0, *fltrPy = 0;
	if ( ! PyArg_ParseTuple( args, "OOO|OOOO", &cfgImpl, &dataseq, &modelsImpl, &rptPy, &rsltPy, &pctPy, &fltrPy ) )
		return NULL;

	if ( (! qubtree_node_check( cfgImpl )) || (! PySequence_Check( dataseq )) || (! qubtree_node_check( modelsImpl )) ) {
		PyErr_SetString( PyExc_TypeError, "MSearch(cfgTree, [idlTree1, idlTree2, ...], modelsTree)" );
		return NULL;
	}

	QUB_Tree cfg = qubtree_node_obj( cfgImpl );

	std::vector<QUB_Tree> data( PySequence_Size( dataseq ) );
	if ( ! data.size() ) {
		PyErr_SetString( PyExc_TypeError, "MIL needs at least one data set" );
		return NULL;
	}
	for ( int i=0; i<int(data.size()); ++i ) {
		PyObject *dat = PySequence_GetItem( dataseq, i );
		if ( qubtree_node_check( dat ) ) {
			data[i] = qubtree_node_obj( dat );
			Py_DECREF( dat );
		}
		else {
			Py_DECREF( dat );
			PyErr_SetString( PyExc_TypeError, "arg 2 must be a sequence of DataSet trees" );
			return NULL;
		}
	}

	QUB_Tree models = qubtree_node_obj( modelsImpl );

	QTR_NodeFptr rptCB, rsltCB, pctCB, fltrCB;
	if ( rptPy )
		rptCB = QTR_NodeFptr( new QTR_NodePyFunc( rptPy ) );
	else
		rptCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( rsltPy )
		rsltCB = QTR_NodeFptr( new QTR_NodePyFunc( rsltPy ) );
	else
		rsltCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( pctPy )
		pctCB = QTR_NodeFptr( new QTR_NodePyFunc( pctPy ) );
	else
		pctCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( fltrPy )
		fltrCB = QTR_NodeFptr( new QTR_NodePyFunc( fltrPy ) );
	else
		fltrCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	QUB_Tree result = msearch_tree( data, cfg, models, rptCB, rsltCB, pctCB, fltrCB );

	return qubtree_node_using( result.getImpl() );
}

string ModelClassOf( QUB_Tree model, std::vector<int>& reorderedStates )
{
	ModelTree mdlTree( model );
	mdlinf mi;
	mdlTree.toMdlinf( mi );
	mdlTree.restoreClasses();

	int i;

	std::vector<int> classSize;
	for ( i=0; i<mi.nstate; ++i ) {
		int cls = mi.clazz[i];
		if ( cls >= int(classSize.size()) )
			classSize.resize( cls + 1 );
		classSize[cls]++;
	}
	std::vector<int> classOrder;
	for ( i=mi.nstate; i>0; --i ) {
		for ( int j=0; j<int(classSize.size()); ++j ) {
			if ( classSize[j] == i )
				classOrder.push_back( j );
		}
	}

	ostrstream mdlClass;
	mdlClass << "N" << mi.nstate;
	reorderedStates.clear();

	for ( i=0; i<int(classOrder.size()); ++i ) {
		for ( int j=0; j<mi.nstate; ++j ) {
			if ( mi.clazz[j] == classOrder[i] ) {
				mdlClass << (i+1);
				reorderedStates.push_back( j );
			}
		}
	}

	mdlClass << '\0';
	string answer( mdlClass.str() );
	mdlClass.rdbuf()->freeze(0);

	return answer;
}

void msearch_modelCount( QUB_Tree model, QUB_Tree models, int *allModelCount, int *noLoopCount )
{
	std::vector<int> reorderedStates; // discarded oh well
	string modelClass = ModelClassOf( model, reorderedStates );
	QUB_Tree modelList = models[ modelClass ];
	
	*allModelCount = modelList.dataCount();
	*noLoopCount   = modelList["NoLoops"].dataCount();
}

extern "C" MIL_API void __stdcall
msearch_modelCount_iface( QTR_Impl *model, QTR_Impl *models, int *allModelCount, int *noLoopCount )
{
	msearch_modelCount( QUB_Tree( model ), QUB_Tree( models ), allModelCount, noLoopCount );
}

